#include <torch/extension.h>
#include <vector>

// Optimized CUDA kernel with reduced warp divergence
__global__ void lstm_elementwise_forward(
    const float* __restrict__ gates,
    const float* __restrict__ prev_c,
    float* __restrict__ h,
    float* __restrict__ c,
    int batch_size,
    int hidden_size
) {
    // Align to warp size (32 threads)
    const int warp_size = 32;
    const int lane_id = threadIdx.x % warp_size;
    const int warp_id = threadIdx.x / warp_size;
    const int warps_per_block = blockDim.x / warp_size;
    const int global_warp_id = blockIdx.x * warps_per_block + warp_id;
    
    // Process elements in warp-aligned chunks
    for (int base_idx = global_warp_id * warp_size; 
         base_idx < batch_size * hidden_size; 
         base_idx += gridDim.x * warps_per_block * warp_size) {
        
        int idx = base_idx + lane_id;
        
        // Pre-compute bounds check for entire warp
        bool valid = idx < batch_size * hidden_size;
        
        if (valid) {
            int b = idx / hidden_size;
            int n = idx % hidden_size;
            int gate_idx = b * hidden_size * 4 + n;

            float i_gate = gates[gate_idx];
            float f_gate = gates[gate_idx + hidden_size];
            float g_gate = gates[gate_idx + 2 * hidden_size];
            float o_gate = gates[gate_idx + 3 * hidden_size];

            // Vectorized activation computations
            i_gate = 1.0f / (1.0f + expf(-i_gate));
            f_gate = 1.0f / (1.0f + expf(-f_gate));
            g_gate = tanhf(g_gate);
            o_gate = 1.0f / (1.0f + expf(-o_gate));

            float c_new = f_gate * prev_c[idx] + i_gate * g_gate;
            float h_new = o_gate * tanhf(c_new);

            c[idx] = c_new;
            h[idx] = h_new;
        }
    }
}

// Optimized linear kernel with reduced warp divergence
__global__ void linear_forward_kernel(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int batch_size,
    int input_dim,
    int output_dim
) {
    const int warp_size = 32;
    const int lane_id = threadIdx.x % warp_size;
    const int warp_id = threadIdx.x / warp_size;
    const int warps_per_block = blockDim.x / warp_size;
    const int global_warp_id = blockIdx.x * warps_per_block + warp_id;
    
    for (int base_idx = global_warp_id * warp_size;
         base_idx < batch_size * output_dim;
         base_idx += gridDim.x * warps_per_block * warp_size) {
        
        int idx = base_idx + lane_id;
        bool valid = idx < batch_size * output_dim;
        
        if (valid) {
            int batch_idx = idx / output_dim;
            int out_idx = idx % output_dim;
            float sum = 0.0f;
            
            // Coalesced memory access pattern
            #pragma unroll 4
            for (int k = 0; k < input_dim; k++) {
                sum += input[batch_idx * input_dim + k] * 
                       weight[out_idx * input_dim + k];
            }
            
            if (bias != nullptr) {
                sum += bias[out_idx];
            }
            output[idx] = sum;
        }
    }
}

torch::Tensor lstm_forward_cuda(
    torch::Tensor input,
    torch::Tensor w_ih,
    torch::Tensor w_hh,
    torch::Tensor b_ih,
    torch::Tensor b_hh,
    torch::Tensor h0,
    torch::Tensor c0
) {
    auto options = torch::TensorOptions().dtype(input.dtype()).device(input.device());
    int batch_size = input.size(0);
    int seq_len = input.size(1);
    int hidden_size = h0.size(1);

    torch::Tensor h = h0.clone();
    torch::Tensor c = c0.clone();
    std::vector<torch::Tensor> outputs;

    const int threads_per_block = 256;
    const int num_warps = threads_per_block / 32;
    const int total_elements = batch_size * hidden_size;
    const int blocks = (total_elements + threads_per_block - 1) / threads_per_block;

    for (int t = 0; t < seq_len; ++t) {
        torch::Tensor xt = input.select(1, t);
        torch::Tensor gates = torch::addmm(b_ih, xt, w_ih.t());
        gates = torch::addmm(gates, h, w_hh.t());
        gates += b_hh;

        lstm_elementwise_forward<<<blocks, threads_per_block>>>(
            gates.data_ptr<float>(),
            c.data_ptr<float>(),
            h.data_ptr<float>(),
            c.data_ptr<float>(),
            batch_size,
            hidden_size
        );

        outputs.push_back(h.unsqueeze(1));
    }

    return torch::cat(outputs, 1);
}

torch::Tensor linear_forward_cuda(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias
) {
    const auto batch_size = input.size(0);
    const auto input_dim = input.size(1);
    const auto output_dim = weight.size(0);

    auto options = torch::TensorOptions().dtype(input.dtype()).device(input.device());
    auto output = torch::empty({batch_size, output_dim}, options);

    const int threads_per_block = 256;
    const int blocks = (batch_size * output_dim + threads_per_block - 1) / threads_per_block;

    linear_forward_kernel<<<blocks, threads_per_block>>>(
        input.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.defined() ? bias.data_ptr<float>() : nullptr,
        output.data_ptr<float>(),
        batch_size,
        input_dim,
        output_dim
    );

    return output;
}

torch::Tensor forward(
    torch::Tensor x,
    std::vector<torch::Tensor> lstm_weights_ih,
    std::vector<torch::Tensor> lstm_weights_hh,
    std::vector<torch::Tensor> lstm_biases_ih,
    std::vector<torch::Tensor> lstm_biases_hh,
    torch::Tensor fc_weight,
    torch::Tensor fc_bias,
    torch::Tensor h0,
    torch::Tensor c0,
    bool is_training
) {
    h0 = h0.to(x.device());
    c0 = c0.to(x.device());
    torch::Tensor out = x;
    
    for (int i = 0; i < lstm_weights_ih.size(); ++i) {
        auto w_ih = lstm_weights_ih[i].to(x.device());
        auto w_hh = lstm_weights_hh[i].to(x.device());
        auto b_ih = lstm_biases_ih[i].to(x.device());
        auto b_hh = lstm_biases_hh[i].to(x.device());
        
        torch::Tensor h_i = h0.narrow(0, i, 1).squeeze(0);
        torch::Tensor c_i = c0.narrow(0, i, 1).squeeze(0);
        
        out = lstm_forward_cuda(out, w_ih, w_hh, b_ih, b_hh, h_i, c_i);
    }
    
    out = out.select(1, -1);
    out = linear_forward_cuda(out, fc_weight, fc_bias);
    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Module forward (CUDA)");
}